use axum::{extract::State, response::IntoResponse};

use crate::{signaling::redisexchange, types::AppState};
use axum::extract::ws::Message;
use std::sync::Arc;
use tokio::sync::{Mutex, RwLock};

pub async fn handle_message_from_frontend(
    mut ws: axum::extract::ws::WebSocket,
    redis: redis::aio::MultiplexedConnection,
    pubsub: Arc<Mutex<redis::aio::PubSub>>,
) {
    let (peer_tx, mut peer_rx) = tokio::sync::mpsc::channel(128);
    let (browser_tx, browser_rx) = tokio::sync::mpsc::channel(128);

    tokio::spawn(redisexchange::send_message_to_peers(
        browser_rx, peer_tx, redis, pubsub,
    ));
    loop {
        tokio::select! {
            msg_from_peer = peer_rx.recv() => {
                let msg = match msg_from_peer {
                    Some(msg) => msg,
                    None => break
                };
                let _ = ws.send(axum::extract::ws::Message::Text(serde_json::to_string(&msg).unwrap())).await;
            },
            msg_from_ws = ws.recv() => {

                let msg = match msg_from_ws {
                    Some(msg) => msg,
                    None => break
                };
                let msg = match msg {
                    Ok(msg) => msg,
                    Err(error) => {
                        tracing::error!(?error, "websocket read error");
                        break;
                    }
                };
                match msg {
                    Message::Text(text) => {
                        let msg = match serde_json::from_str(&text) {
                            Ok(msg) => msg,
                            Err(error) => {
                                tracing::warn!(?error, "failed to decode websocket");
                                continue;
                            },
                        };

                        let _ = browser_tx.send(msg).await;
                    }
                    _ => continue,
                }
            }
        }
    }
    // browser_rx.close();
}

pub async fn handle_ws_upgrade(
    upgrade: axum::extract::WebSocketUpgrade,
    State(state): State<Arc<RwLock<AppState>>>,
) -> Result<impl IntoResponse, String> {
    let state = state.read().await;
    let redis: redis::aio::MultiplexedConnection =
        match state.redis.get_multiplexed_tokio_connection().await {
            Ok(redis) => redis,
            Err(err) => return Err(err.to_string()),
        };

    let pubsub = match state.clone().redis.get_async_pubsub().await {
        Ok(pubsub) => pubsub,
        Err(err) => return Err(err.to_string()),
    };

    Ok(upgrade
        .on_upgrade(|ws| handle_message_from_frontend(ws, redis, Arc::new(Mutex::new(pubsub)))))
}
