use std::sync::Arc;

use axum::{
    extract::State,
    response::{ErrorResponse, Html},
    routing::{get, post, IntoMakeService},
    Json, Router,
};
use serde::Serialize;
use tokio::sync::RwLock;
use tower_http::trace::TraceLayer;

use crate::{
    signaling::grpc::SignalingService as GrpcSignalingService,
    types::{AppState, PageContext},
    web::templates::load_templates,
};

#[cfg(feature = "serve-static")]
use crate::web::staticfiles;

async fn root(State(state): State<Arc<RwLock<AppState>>>) -> axum::response::Result<Html<String>> {
    let state = state.read().await;

    let ctx = match tera::Context::from_serialize(&PageContext::default()) {
        Ok(ctx) => ctx,
        Err(err) => return Err(ErrorResponse::from(format!("{err}"))),
    };
    match state.templates.render("index.html", &ctx) {
        Ok(result) => Ok(Html::from(result)),
        Err(err) => Err(ErrorResponse::from(format!("{err}"))),
    }
}

#[derive(Serialize, Default)]
struct ReloadResult {
    #[serde(skip_serializing_if = "Option::is_none")]
    error: Option<String>,
}

async fn reload(State(state): State<Arc<RwLock<AppState>>>) -> Json<ReloadResult> {
    let mut state = state.write_owned().await;
    if let Err(err) = state.templates.full_reload() {
        return Json(ReloadResult {
            error: Some(err.to_string()),
        });
    }
    drop(state);

    Json(ReloadResult::default())
}

pub struct Config {
    pub redis_addr: String,
}

impl Default for Config {
    fn default() -> Self {
        Self {
            redis_addr: String::from("redis://127.0.0.1")
        }
    }
}


pub fn make_service(config: Option<Config>) -> Result<IntoMakeService<Router<()>>, Box<dyn std::error::Error>> {
    let config = config.unwrap_or_default();
    let templates = load_templates()?;

    let router = Router::new();

    #[cfg(feature = "serve-static")]
    let router = router.route_service("/static/*path", staticfiles::StaticFiles::strip("/static/"));

    let redis = redis::Client::open(config.redis_addr)?;

    let grpc_service = tonic_web::enable(signaling::signaling_server::SignalingServer::new(
        GrpcSignalingService {
            redis: redis.clone(),
        },
    ));

    Ok(router
        .route("/", get(root))
        .route("/reload", post(reload))
        .route(
            "/signaling.Signaling/*rpc",
            axum::routing::any_service(grpc_service.clone()),
        )
        .route("/ws", get(crate::signaling::websocket::handle_ws_upgrade))
        .with_state(Arc::new(RwLock::new(crate::types::AppState {
            templates: templates,
            redis,
            // names: Arc::new(Mutex::new(names::Generator::default())),
        })))
        .layer(TraceLayer::new_for_http())
        // .into_make_service_with_connect_info()
        .into_make_service())
}
